#include <stdio.h>
#include <stdlib.h>         /* calloc */
#include <netinet/in.h>     
#include <netinet/ip.h>     /* struct iphdr */
#include <netdb.h>          /* /etc/protocols */
#include <unistd.h>         /* read */
#include <sys/stat.h>       /* open */
#include <arpa/inet.h>      /* inet_pton */
#include <string.h>         /* strerror */
#include <sys/socket.h>
#include <fcntl.h>          /* O_RDONLY */

#include "ospf.h"

/* http://www.tenouk.com/Module43a.html */
#define PCKT_LEN 8192

/* Need a random value for the Ident field 
   Kinda neat...http://stackoverflow.com/a/7455622 
   Even more interesting - looks like Linux won't let a zero out.
   If you leave this empty, Linux will assign a random value anyways.
*/
uint16_t randomIdentifier ()
{
  int myFile;
  uint16_t id;
  
  myFile = open("/dev/random", O_RDONLY);
  read(myFile, &id, sizeof(uint16_t));
  close(myFile);

  return htons(0xdead);
}

void initIP(struct iphdr * ip)
{
  int myFile, status;

  ip->version = 4;                 /* IPv4 */
  ip->ihl     = 5;                 /* No options, header length = 20 bytes */
  ip->tos     = IPTOS_CLASS_CS6;   /* from netinet/ip.h */
  ip->tot_len = UINT16_MAX;        /* initialize to max value */

  /* Need a random value for the Ident field */
  /* Kinda neat...http://stackoverflow.com/a/7455622 */
  myFile = open("/dev/random", O_RDONLY);
  read(myFile, &(ip->id), sizeof(uint16_t));
  close(myFile);

  ip->id = randomIdentifier();
  
  ip->frag_off = 0;      /* fragmentation is so not cool anymore */
  
  /* From http://tools.ietf.org/html/rfc2328#appendix-A.1
     "To ensure that these packets will not travel multiple hops,
     their IP TTL must be set to 1." */
  ip->ttl = 1;

  /* http://www.iana.org/assignments/protocol-numbers */
  ip->protocol = getprotobyname("ospf")->p_proto; /* I know it's 89... */

  ip->check = UINT16_MAX;               /* initialize to max value */

  /* Source IP Address */
  if ((status = inet_pton (AF_INET, "10.0.10.11", &(ip->saddr))) != 1)
    {
      fprintf(stderr, "inet_pton error: %s\n", strerror(status));
      exit(EXIT_FAILURE);
    }

  /* Destination IP Address */
  if ((status = inet_pton (AF_INET, "224.0.0.5", &(ip->daddr))) != 1)
    {
      fprintf(stderr, "inet_pton error: %s\n", strerror(status));
      exit(EXIT_FAILURE);
    }
}

void printIP (struct iphdr * ip)
{
  char saddr [INET_ADDRSTRLEN], daddr [INET_ADDRSTRLEN];
  inet_ntop (AF_INET, &(ip->saddr), saddr, INET_ADDRSTRLEN);
  inet_ntop (AF_INET, &(ip->daddr), daddr, INET_ADDRSTRLEN);

  printf("Internet Protocol Version 4\n");
  printf("\tVersion: %d\n", ip->version);
  printf("\tHeader Length: %d\n", ip->ihl);
  printf("\tDifferentiated Services Field: %#02x\n", ip->tos);
  printf("\tTotal Length: %d\n", ip->tot_len);
  printf("\tIdentification: %#04x\n", ntohs(ip->id));
  printf("\tFlags: %#02x\n", ip->frag_off);
  printf("\tFragment Offset: %d\n", ip->frag_off);
  printf("\tTime to live: %d\n", ip->ttl);
  printf("\tProtocol: %s (%d)\n", getprotobynumber(ip->protocol)->p_aliases[0],
	 ip->protocol);
  printf("\tHeader checksum: %#04x\n", ip->check);
  printf("\tSource: %s\n", saddr);
  printf("\tDestination: %s\n", daddr);
}

void initOSPFHeader(struct OSPFheader * header)
{
  header->version = 2;
  header->msgtype = 1;
  /* Skip length */
  header->rid     = htonl(1);           /* 0.0.0.1 */
  header->areaid  = 0;
  /* Skip checksum */
  header->autype  = 0;
  header->authen  = 0;
}

const char * getOSPFMessageType (unsigned short msgtype)
{
  const char * r;
  const char * messageTypes [] = {
    "Hello Packet",
    "Database Description",
    "Link State Request",
    "Link State Update",
    "Link State Acknowledgement"
  };
  const char * invalid = "Invalid";
  
  if (msgtype > 5 || msgtype == 0)
    r = invalid;
  else
    r = messageTypes[msgtype - 1];
  return r;
}

const char * getOSPFAuthType (unsigned short autype)
{
  const char * r;
  const char * authTypes [] = {
    "Null",
    "Simple password",
    "Cryptographic authentication"
  };
  const char * invalid = "Reserved for assignment by the IANA (iana@ISI.EDU)";

  if (autype > 2)
    r = invalid;
  else
    r = authTypes[autype];
  return r;
}

void printOSPFHeader(struct OSPFheader * header)
{
  char rid [INET_ADDRSTRLEN], areaid [INET_ADDRSTRLEN];
  inet_ntop (AF_INET, &(header->rid), rid, INET_ADDRSTRLEN);
  inet_ntop (AF_INET, &(header->areaid), areaid, INET_ADDRSTRLEN);
  
  printf("Open Shortest Path First\n");
  printf("\tVersion: %d\n", header->version);
  printf("\tMessage Type: %s (%d)\n",
	 getOSPFMessageType(header->msgtype), header->msgtype);
  printf("\tPacket Length: %d\n", ntohs(header->length));
  printf("\tSource OSPF Router: %s (%s)\n", rid, rid);
  printf("\tArea ID: %s (%s)", areaid, areaid);
  if (header->areaid == 0)
    printf(" (Backbone)");
  printf("\n");
  printf("\tChecksum: %#04x\n", ntohs(header->chksum));
  printf("\tAuth Type: %s (%d)\n", getOSPFAuthType(ntohs(header->autype)),
	 ntohs(header->autype));
  printf("\tAuth Data");
  if (header->authen == 0)
    printf(" (none)");
  printf(": %016llx", (long long unsigned int)header->authen);
  if (header->authen != 0)
    printf("(warning: byte order not corrected!)");
  printf("\n");
}

void initOSPFHello(struct OSPFhello * hello)
{
  hello->netmask = htonl(0xffffff00);   /* 255.255.255.0 */
  hello->hellointerval = htons(10);
  hello->options  = 0;                  /* All options turned off */
  hello->priority = 0;
  hello->deadint  = htonl(40);
  hello->dr  = 0;                       /* What if priority = 0 but dr = my IP? */
  hello->bdr = 0;
}

int main ()
{
  struct iphdr * ip;
  struct OSPFheader * header;
  struct OSPFhello * hello;
  
  uint8_t buffer[PCKT_LEN];
  int sd;
  struct sockaddr_in sin;
  const int on = 1;

  memset(buffer, 0, PCKT_LEN);

  ip = (struct iphdr *) buffer;
  header = (struct OSPFheader *) (buffer + sizeof(struct iphdr));
  hello = (struct OSPFhello *) (buffer + sizeof(struct iphdr) +
			        sizeof(struct OSPFheader));

  initIP(ip);
  initOSPFHeader(header);
  initOSPFHello(hello);
  
  ip->tot_len = sizeof(struct iphdr) + sizeof(struct OSPFheader) +
    sizeof(struct OSPFhello); /* single-byte field, so no htons */

  header->length = htons(sizeof(struct OSPFheader) + sizeof(struct OSPFhello));

  ip->check = 0;
  ip->check = chksum(ip, ip->ihl * 4); // interesting fact: do not change byte ordering

  header->chksum = 0;
  header->chksum = chksum(header, ntohs(header->length));

  printIP(ip);
  printOSPFHeader(header);
  
  sin.sin_family = AF_INET;
  sin.sin_addr.s_addr = ip->daddr;

  sd = socket(PF_INET, SOCK_RAW, IPPROTO_RAW);
  if (sd < 0)
    {
      perror("socket() error");
      exit(EXIT_FAILURE);
    }

  if (setsockopt(sd, IPPROTO_IP, IP_HDRINCL, &on, sizeof(on)) < 0)
    {
      perror("setsockopt() error");
      exit(EXIT_FAILURE);
    }

  if (sendto(sd, buffer, ip->tot_len, 0, (struct sockaddr *)&sin,
	     sizeof(struct sockaddr)) < 0)
    {
      perror("sendto() failed");
      exit(EXIT_FAILURE);
    }
  
  close(sd);
  return(EXIT_SUCCESS);
}
